from pathlib import Path

import pytest

from addon_odoo_wheel import builder


@pytest.fixture
def addon1_sdist_builder(project2: Path) -> builder.AddonBuilder:
    return builder.AddonBuilder(project2, {})


@pytest.fixture
def module_data() -> Path:
    return Path(__file__).parent / "data"


@pytest.fixture(scope="session")
def data_path() -> Path:
    return Path(__file__).parent.joinpath("data")


@pytest.fixture(scope="session")
def project0(data_path: Path) -> Path:
    return data_path / "project0"


@pytest.fixture(scope="session")
def project1(data_path: Path) -> Path:
    return data_path / "project1"


@pytest.fixture(scope="session")
def project2(data_path: Path) -> Path:
    return data_path / "project2"


@pytest.fixture(scope="session")
def project3(data_path: Path) -> Path:
    return data_path / "project3"



from odoo import api, fields, models


class Users(models.Model):
    _inherit = "res.users"

    awesome_user = fields.Boolean(string="Awesome User")

    @api.model
    def create(self, vals_list):
        for val in vals_list:
            val["awesome_user"] = val.get("login") == "awesome"
        return super().create(vals_list)



import sys

__all__ = ["importlib_metadata", "tomllib", "long_description", "make_classifiers"]

if sys.version_info < (3, 8):  # pragma: no cover (<PY38)
    import importlib_metadata
else:  # pragma: no cover (PY38+)
    import importlib.metadata as importlib_metadata

if sys.version_info < (3, 11):
    import tomli as tomllib
else:
    import tomllib


def long_description(addon):
    from manifestoo_core.metadata import _long_description as manifestoo_long_description

    return manifestoo_long_description(addon)


def make_classifiers(odoo_series, manifest):
    from manifestoo_core.metadata import _make_classifiers as manifestoo_make_classifiers

    return manifestoo_make_classifiers(odoo_series, manifest)



from __future__ import annotations

import email
import logging
import re
from email.message import Message
from email.utils import formataddr
from pathlib import Path

import manifestoo_core.exceptions
from manifestoo_core.addon import Addon
from manifestoo_core.exceptions import UnsupportedManifestVersion
from manifestoo_core.manifest import Manifest
from manifestoo_core.metadata import OdooSeriesInfo
from manifestoo_core.odoo_series import MIN_VERSION_PARTS, OdooSeries
from packaging.requirements import Requirement

from . import compat
from .contributors_api import ContributorFactory
from .utils import (
    EXTRA_NAME_RE,
    WHEEL_TAG,
    PyProjectConfig,
    load_pyproject_toml,
    normalize_label,
    odoo_licence_to_SDPX,
)

_logger = logging.getLogger(__name__)

MANIFEST_VERSION_REGEX = re.compile(
    r"^((?P<odoo_version>\d{2}\.[0-1])\.)?(?P<major>0|[1-9]\d*)(\.(?P<minor>0|[1-9]\d*))?(\.(?P<patch>0|[1-9]\d*))?$"
)


class AddonMetadata:
    addon: Addon
    odoo_series: OdooSeries | None
    odoo_series_info: OdooSeriesInfo | None
    """The Odoo version found from the module version or the key `odoo_version_override` in pyproject.toml"""
    git_postversion: str

    @classmethod
    def from_addon_dir(
        cls, addon_dir: Path, allow_not_installable: bool = True, force_addon_name: str = None
    ) -> "AddonMetadata":
        pyproject_options = load_pyproject_toml(addon_dir)
        try:
            _addon = Addon.from_addon_dir(pyproject_options.addon_src, allow_not_installable)
        except manifestoo_core.exceptions.AddonNotFoundNoManifest:
            if pyproject_options:
                module_name = pyproject_options.name.split("-")[-1]
            if (pyproject_options.addon_src / module_name).exists() and (
                pyproject_options.addon_src / module_name
            ).is_dir():
                _addon = Addon.from_addon_dir(pyproject_options.addon_src / module_name, allow_not_installable)
            else:
                msg = f"No manifest file found in {pyproject_options.addon_src} or in {module_name}"
                raise manifestoo_core.exceptions.AddonNotFoundNoManifest(msg) from None

        if force_addon_name:
            _addon.name = force_addon_name
        return cls(addon_dir, pyproject_options, _addon)

    def __init__(self, addon_dir, pyproject_options: PyProjectConfig, addon: Addon) -> None:
        self._build_path = addon_dir
        self.addon = addon
        self.pyproject_options = pyproject_options
        self.odoo_series = self._get_odoo_series()
        _logger.info("Odoo Series %s", self.odoo_series)
        self.module_contributor = ContributorFactory.from_author(self.manifest.author)
        _logger.info("Contributor %s", self.module_contributor.names[0])
        self.git_postversion = (
            self.pyproject_options.addon_odoo_wheel.post_version_strategy_override
            or self.module_contributor.git_postversion_strategy(self.odoo_series)
        )
        self._pk_name: str | None = None
        self._pk_version: str | None = None

    @property
    def python_addon_path(self) -> str:
        return f"odoo.addons.{self.addon.name}"

    @property
    def addon_path(self):
        return self.addon.path

    @property
    def build_path(self):
        return self._build_path

    @property
    def odoo_series_info(self) -> OdooSeriesInfo | None:
        return self.odoo_series and OdooSeriesInfo.from_odoo_series(self.odoo_series)

    @property
    def wheel_name(self) -> str:
        return f"{self.sdist_name}-{WHEEL_TAG}"

    @property
    def sdist_name(self) -> str:
        return "{}-{}".format(self.pkg_name.replace("-", "_").replace(".", "_"), self.pkg_version)

    @property
    def addon_name(self):
        return self.addon.name

    @property
    def manifest(self) -> Manifest:
        return self.addon.manifest

    def _get_odoo_series(self) -> OdooSeries | None:
        if self.pyproject_options.addon_odoo_wheel.odoo_version_override:
            return OdooSeries.from_str(self.pyproject_options.addon_odoo_wheel.odoo_version_override)
        _version_pyproject = self.pyproject_options.version
        _version_manifest = self.addon.manifest.version
        if not _version_pyproject and not _version_manifest:
            raise ValueError("Manifest version or pyproject.toml version must be set")
        if _version_pyproject and _version_manifest and _version_pyproject != _version_manifest:
            msg = (
                "Manifest version or pyproject.toml version must be the same, "
                "or set version in pyproject.toml to dynamic"
            )
            raise ValueError(msg)
        _version = _version_manifest or _version_pyproject
        matchs = MANIFEST_VERSION_REGEX.match(_version)
        if not matchs:
            msg = "Version in manifest or in your pyproject.toml must follow semver convention:"
            raise UnsupportedManifestVersion(msg)
        odoo_version = matchs.group("odoo_version")
        if odoo_version:
            return OdooSeries.from_str(odoo_version)
        return None

    def generate_metadata(self) -> Message:
        """Return Python Package Metadata 2.1 for an Odoo addon directory as an
        ``email.message.Message``.

        The Description field is absent and is stored in the message payload. All values are
        guaranteed to not contain newline characters, except for the payload.

        ``precomputed_metadata_path`` may point to a file containing pre-computed metadata
        that will be used to obtain the Name and Version, instead of looking at the addon
        directory name or manifest version + VCS, respectively. This is useful to process a
        manifest from a sdist tarball with PKG-INFO, for example, when the original
        directory name or VCS is not available to compute the package name and version.

        This function may raise :class:`manifestoo_core.exceptions.ManifestooException` if
        ``addon_dir`` does not contain a valid installable Odoo addon for a supported Odoo
        version.
        """
        if (self.addon.path / "PKG-INFO").exists():
            # if PKG-INFO is present, assume we are in an sdist, copy everything
            with (self.addon.path / "PKG-INFO").open("r") as f:
                _logger.debug("PKG_INFO file found, using it")
                return email.parser.HeaderParser().parse(f)
        return self.construct_metadata_file_2_4()

    def construct_metadata_file_2_4(self) -> Message:
        """
        https://peps.python.org/pep-0639/
        """
        msg = Message()

        msg["Metadata-Version"] = "2.4"
        msg["Name"] = self.pkg_name
        msg["Version"] = self.pkg_version
        self._set_urls(msg)
        self._set_authors(msg)
        self._set_license(msg)
        self._set_classifier(msg)
        self._set_require_python(msg)
        self._set_dependencies(msg)
        self._set_extra_dependencies(msg)
        self._set_description(msg)
        return msg

    @property
    def pkg_name(self):
        if self._pk_name is not None:
            return self._pk_name

        if self.pyproject_options and not self.pyproject_options.name:
            _logger.error("The `project.name` field must be set in pyproject.toml")
            raise ValueError("The `project.name` field must be set in pyproject.toml")

        project_pkg_name = self.pyproject_options.name
        prefix = self.pyproject_options.addon_odoo_wheel.get("package_prefix")
        if not prefix:
            prefix = self.module_contributor.get_package_prefix(self.odoo_series)
        auto_pkg_name = f"{prefix}-{self.addon_name}"

        if project_pkg_name:
            if self.pyproject_options.addon_odoo_wheel.metadata_src_manifest and auto_pkg_name != project_pkg_name:
                raise ValueError(f"The `project.name` must be `{auto_pkg_name}`")

        self._pk_name = project_pkg_name or auto_pkg_name

        return self._pk_name

    @property
    def pkg_version(self):
        if self._pk_version is not None:
            return self._pk_version

        _version = self.manifest.version
        if "version" not in self.pyproject_options.dynamic and self.pyproject_options.version:
            if _version != self.pyproject_options.version:
                raise ValueError("Version must be same between pyproject.toml and __manifest__.py")
            _logger.debug("Force set version from pyproject.toml: %s -> %s", _version, self.pyproject_options.version)
            self._pk_version = self.pyproject_options.version
            return self._pk_version

        _logger.debug("Manifest version: %s", _version)
        _logger.debug("Odoo version: %s", self.odoo_series and self.odoo_series.value or "undefined")
        _logger.debug("version: %s", _version)
        matches = MANIFEST_VERSION_REGEX.match(_version)
        if not matches:
            msg = f"Can't match version '{_version}', see regex https://regex101.com/r/HjHXm7/1"
            raise UnsupportedManifestVersion(msg)

        odoo_version = matches.group("odoo_version") or None
        pkg_major = matches.group("major") or "1"
        pkg_minor = matches.group("minor") or "0"
        pkg_patch = matches.group("patch") or "0"
        if not odoo_version and not self.odoo_series and self.pyproject_options.addon_odoo_wheel.allow_no_odoo_version:
            self._pk_version = _version
            return self._pk_version
        if not odoo_version and self.odoo_series.value == ".".join((pkg_major, pkg_minor)):
            _logger.debug("Manifest version is same as Odoo version. Force set to 1.0.0")
            odoo_version = ".".join((pkg_major, pkg_minor))
            pkg_major = "1"
            pkg_minor = "0"
            pkg_patch = "0"

        if odoo_version and self.odoo_series:
            if odoo_version != self.odoo_series:
                msg = f"The version in __manifest__.py start with {odoo_version} but {self.odoo_series.value} is found"
                raise UnsupportedManifestVersion(msg)
        if not odoo_version and self.odoo_series:
            _logger.debug(
                "Force set odoo version before module version from odoo_version_override: %s.%s",
                self.odoo_series.value,
                _version,
            )
            odoo_version = self.odoo_series.value

        _logger.debug(matches)
        normal_version = ".".join(
            tuple(
                filter(
                    bool,
                    (
                        odoo_version,
                        pkg_major,
                        pkg_minor,
                        pkg_patch,
                    ),
                )
            )
        )
        matches: re.Match = MANIFEST_VERSION_REGEX.match(normal_version)
        if not matches:
            msg = f"Can't match version '{normal_version}', see regex https://regex101.com/r/HjHXm7/1"
            raise UnsupportedManifestVersion(msg)
        self._pk_version = normal_version
        _logger.debug("Normalized version -> %s", normal_version)
        return self._pk_version

    def _set_description(self, msg: Message):
        if self.pyproject_options.get("summary"):
            msg["Summary"] = self.pyproject_options.get("summary").splitlines()[0]
        to_try = [("README.md", "text/markdown"), ("README.rst", "text/x-rst")]
        if self.pyproject_options.get("readme"):
            content_type = "text/plain"
            if Path(self.pyproject_options.get("readme")).suffix == ".md":
                content_type = "text/markdown"
            if Path(self.pyproject_options.get("readme")).suffix == ".rst":
                content_type = "text/x-rst"
            to_try.insert(0, (self.pyproject_options.get("readme"), content_type))

        for read_to_try in to_try:
            if not read_to_try[0]:
                continue
            readme_path = self.addon.path / read_to_try[0]
            if readme_path.exists() and readme_path.is_file():
                msg["Description-Content-Type"] = read_to_try[1]
                msg.set_payload(readme_path.read_text(encoding="utf-8"))
                break
        else:
            if self.manifest.description:
                msg["Description-Content-Type"] = "text/plain"
                msg["Description"] = self.manifest.description

    def _set_dependencies(self, msg: Message):
        dependencies = []
        source = {
            "project.dependencies" if self.pyproject_options.dependencies else None,
            "addon_odoo_wheel.dependencies" if self.pyproject_options.addon_odoo_wheel.dependencies else None,
            "addon_odoo_wheel.dependencies_file" if self.pyproject_options.addon_odoo_wheel.dependencies_file else None,
        }
        source.remove(None)
        if len(source) > 1:
            msg = "Only one source of 'dependencies' can be set. Found: " + ", ".join(source)
            raise ValueError(msg)
        if len(source) == 0 and (self.build_path / "requirements.txt").exists():
            _logger.debug("Using 'requirements.txt' inside %s for dependencies", self.build_path)
            self.pyproject_options.addon_odoo_wheel["dependencies-file"] = self.build_path / "requirements.txt"

        if self.pyproject_options.dependencies:
            dependencies = self.pyproject_options.dependencies
        if self.pyproject_options.addon_odoo_wheel.dependencies:
            dependencies = self.pyproject_options.addon_odoo_wheel.dependencies
        if self.pyproject_options.addon_odoo_wheel.dependencies_file:
            with self.pyproject_options.addon_odoo_wheel.dependencies_file.open("r") as f:
                dependencies = [Requirement(dep) for dep in f.read().splitlines() if not dep.startswith("#")]

        for dependency in dependencies:
            msg["Requires-Dist"] = str(dependency)

    def _set_extra_dependencies(self, msg: Message):
        extras = self.pyproject_options.get("optional-dependencies") or self.pyproject_options.addon_odoo_wheel.get(
            "optional-dependencies"
        )
        if not extras:
            return

        for extra, dependencies in extras.items():
            if not EXTRA_NAME_RE.match(extra):
                raise ValueError(
                    f"Invalid extra name: {extra!r} See: https://packaging.python.org/en/latest/specifications/core-metadata/#provides-extra-multiple-use"
                )
            msg["Provides-Extra"] = extra
            for dependency in dependencies:
                msg["Requires-Dist"] = f"{dependency}; extra == {extra!r}"

    def _set_require_python(self, msg: Message):
        requires_python = (
            self.pyproject_options.requires_python
            or (self.odoo_series and self.odoo_series_info.python_requires)
            or self.pyproject_options.addon_odoo_wheel.requires_python
        )
        if requires_python:
            msg["Requires-Python"] = requires_python

    def _set_classifier(self, msg: Message):
        classifiers = set(self.pyproject_options.classifiers)
        if not self.pyproject_options:
            classifiers |= set(self.pyproject_options.addon_odoo_wheel.classifiers)
        classifiers |= set(self.module_contributor.get_classifiers(self.odoo_series))
        _, classifier_license = odoo_licence_to_SDPX.get(self.manifest.license, (None, None))
        if classifier_license:
            classifiers.add(classifier_license)
        if self.odoo_series:
            _manifest_copy = Manifest.from_dict(self.manifest.manifest_dict)
            if not self.manifest.development_status:
                version_parts = self.manifest.version.split(".")
                if len(version_parts) == MIN_VERSION_PARTS:
                    version_parts = version_parts[2:]
                development_status = "stable" if int(version_parts[0]) > 0 else "alpha"
                _manifest_copy = Manifest.from_dict(
                    dict(self.manifest.manifest_dict, development_status=development_status)
                )
            classifiers |= set(compat.make_classifiers(self.odoo_series, _manifest_copy))
        for classifier in sorted(list(classifiers)):
            msg["Classifier"] = classifier

    def _set_license(self, msg: Message):
        license_map, _ = odoo_licence_to_SDPX.get(self.manifest.license, (None, None))
        if self.addon.manifest.license == "Other proprietary" and not self.pyproject_options.get("license"):
            raise ValueError("field `project.license` must be static when using 'Other proprietary' value in manifest")
        msg["License-Expression"] = license_map or self.pyproject_options.get("license") or "AGPL-3.0-or-later"
        for license_file in self.pyproject_options.get("license_files") or []:
            msg["License-File"] = license_file

    def _set_urls(self, msg: Message):
        urls = self.pyproject_options.get("urls", {})
        urls["Homepage"] = self.manifest.website
        for label, url in urls.items():
            is_well_know, label_n = normalize_label(label)
            msg["Project-URL"] = f"{label_n}, {url}"

    def _set_authors(self, msg: Message):
        odoo_authors = []
        if self.manifest.author:
            odoo_authors = self.manifest.author.split(",")
        author_values = []
        author_no_mail = []
        for author in self.pyproject_options.get("authors", []):
            if author["name"] in odoo_authors:
                odoo_authors.remove(author["name"])
            author_values.append(formataddr((author["name"], author["email"])))

        for odoo_author in odoo_authors:
            email = self.module_contributor.get_mail(self.odoo_series)
            if email:
                author_values.append(formataddr((odoo_author, email)))
            else:
                author_no_mail.append(odoo_author)
        if author_values:
            msg["Author-Email"] = ", ".join(author_values)
        if author_no_mail:
            msg["Author"] = ", ".join(author_no_mail)


def _no_nl(s: str | None) -> str | None:
    if not s:
        return s
    return " ".join(s.split())



from __future__ import annotations

import configparser
import logging
import os
import re
import shutil
import string
import subprocess
import typing
from email.generator import Generator
from email.message import Message
from pathlib import Path
from typing import Any, Callable

from manifestoo_core.metadata import EXTERNAL_DEPENDENCIES_MAP
from packaging.requirements import Requirement
from typing_extensions import TypeAlias

from .compat import tomllib
from .exceptions import NoScmFound
from .version import version as lib_version

_logger = logging.getLogger(__name__)
if typing.TYPE_CHECKING:
    pass


def _get_mapping_metadata():
    return {"-".join(t.capitalize() for t in k.split("-")): v for k, v in EXTERNAL_DEPENDENCIES_MAP.items()}


EXTRA_NAME_RE = re.compile("^[a-z0-9]+(-[a-z0-9]+)*$")

CORE_METADATA_PROJECT_FIELDS = {
    "Author": ("authors",),
    "Author-email": ("authors",),
    "Classifier": ("classifiers",),
    "Description": ("readme",),
    "Description-Content-Type": ("readme",),
    "Dynamic": ("dynamic",),
    "Keywords": ("keywords",),
    "License": ("license",),
    "License-Expression": ("license",),
    "License-Files": ("license-files",),
    "Maintainer": ("maintainers",),
    "Maintainer-email": ("maintainers",),
    "Name": ("name",),
    "Provides-Extra": ("dependencies", "optional-dependencies"),
    "Requires-Dist": ("dependencies",),
    "Requires-Python": ("requires-python",),
    "Summary": ("description",),
    "Project-URL": ("urls",),
    "Version": ("version",),
}
PROJECT_CORE_METADATA_FIELDS = {
    "authors": ("Author", "Author-email"),
    "classifiers": ("Classifier",),
    "dependencies": ("Requires-Dist",),
    "dynamic": ("Dynamic",),
    "keywords": ("Keywords",),
    "license": ("License", "License-Expression"),
    "license-files": ("License-Files",),
    "maintainers": ("Maintainer", "Maintainer-email"),
    "name": ("Name",),
    "optional-dependencies": ("Requires-Dist", "Provides-Extra"),
    "readme": ("Description", "Description-Content-Type"),
    "requires-python": ("Requires-Python",),
    "description": ("Summary",),
    "urls": ("Project-URL",),
    "version": ("Version",),
}
WHEEL_TAG = "py3-none-any"

"""
Mapping with SDPX licence from https://spdx.org/licenses/
"""
odoo_licence_to_SDPX = {
    "GPL-2": ("GPL-2.0", "License :: OSI Approved :: GNU General Public License v2 (GPLv2)"),
    "GPL-2 or any later version": (
        "GPL-2.0-or-later",
        "License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)",
    ),
    "GPL-3": ("GPL-3.0-only", "License :: OSI Approved :: GNU General Public License v3 (GPLv3)"),
    "GPL-3 or any later version": (
        "GPL-3.0-or-later",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
    ),
    "AGPL-3": (
        "AGPL-3.0",
        "License :: OSI Approved :: GNU Affero General Public License v3",
    ),
    "AGPL-3 or any later version": (
        "AGPL-3.0",
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
    ),
    "LGPL-2": ("LGPL-2.0", "License :: OSI Approved :: GNU Lesser General Public License v2 (LGPLv2)"),
    "LGPL-2 or any later version": (
        "LGPL-2.0",
        "License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)",
    ),
    "LGPL-3": ("LGPL-3", "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)"),
    "LGPL-3 or any later version": (
        "LGPL-3.0-or-later",
        "License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)",
    ),
    "Other OSI approved licence": ("any-OSI", None),
    "OEEL-1": ("LicenseRef-OEEL-1", None),
    "OPL-1": ("LicenseRef-OPL-1", None),
    "Other proprietary": (None, None),
}


def base_wheel_metadata() -> Message:
    """
    Create and return the base metadata for the wheel.

    This metadata includes the wheel version, the generator information,
    whether the root is purelib, and the tag. It is used in the creation
    of the wheel file for the Odoo addon.

    Returns:
        Message: An email.message.Message object containing the base wheel metadata.
    """
    msg = Message()
    msg["Wheel-Version"] = "1.0"  # of the spec
    msg["Generator"] = "Mangono Wheel Builder" + lib_version
    msg["Root-Is-Purelib"] = "true"
    msg["Tag"] = WHEEL_TAG
    return msg


def write_metadata(msg: Message, dest: Path) -> None:
    with open(dest, "w", encoding="utf-8") as f:
        Generator(f, mangle_from_=False, maxheaderlen=0).flatten(msg)


EntryPoints: TypeAlias = typing.Dict[str, typing.Dict[str, str]]


class CaseSensitiveConfigParser(configparser.ConfigParser):
    optionxform = staticmethod(str)


def create_entry_point(entry_points: EntryPoints, python_addon_path: str, dest: Path):
    if not entry_points:
        return
    entrypoint_format = CaseSensitiveConfigParser()
    for entry_key, entry_value in entry_points.items():
        entrypoint_format.add_section(entry_key)
        for entry_name, entry_path in entry_value.items():
            odoo_entry_path = f"{python_addon_path}.{entry_path}"
            if entry_path.startswith(python_addon_path):
                odoo_entry_path = entry_path
            entrypoint_format.set(entry_key, entry_name, odoo_entry_path)
    entrypoint_format.write((dest / "entry_points.txt").open("w"))


def load_pyproject_toml(addon_dir: Path) -> PyProjectConfig:
    pyproject_toml_path = addon_dir / "pyproject.toml"
    if pyproject_toml_path.exists():
        with open(pyproject_toml_path, "rb") as f:
            return PyProjectConfig.from_config(addon_dir, tomllib.load(f))
    return PyProjectConfig.from_config(addon_dir, {})


WellKnownLabels = [
    "homepage",
    "source",
    "download",
    "changelog",
    "releasenotes",
    "documentation",
    "issues",
    "funding",
]


def normalize_label(label: str) -> tuple[bool, str]:
    chars_to_remove = string.punctuation + string.whitespace
    removal_map = str.maketrans("", "", chars_to_remove)
    n_label = label.translate(removal_map).lower()
    return n_label in WellKnownLabels, n_label.capitalize()


class PyProjectConfigBase(dict):
    @classmethod
    def from_file(cls, pyproject_toml_path: Path) -> PyProjectConfig:
        if not pyproject_toml_path.exists():
            raise FileNotFoundError(pyproject_toml_path)
        with open(pyproject_toml_path, "rb") as f:
            return cls.from_config(tomllib.load(f))

    @property
    def dependencies(self) -> list[Requirement]:
        return [Requirement(dep) for dep in self.get("dependencies") or []]

    @property
    def optional_dependencies(self) -> dict[str, list[Requirement]]:
        return {
            option: [Requirement(dep) for dep in deps] for option, deps in self.get("optional-dependencies", {}).items()
        }

    @property
    def author(self) -> str | None:
        return self.get("author")

    @property
    def author_email(self) -> str | None:
        return self.get("author_email")

    @property
    def requires_python(self) -> str | None:
        return self.get("requires_python")

    @property
    def classifiers(self) -> list[str]:
        return self.get("classifiers", [])

    @property
    def entry_points(self) -> EntryPoints:
        return self.get("entry-points", {})


class ToolConfig(PyProjectConfigBase):
    @property
    def support_only_community(self) -> bool:
        return bool(self.get("only-community") or True)

    @property
    def post_version_strategy_override(self) -> str | None:
        post_version_strategy_override = self.get("post_version_strategy_override")
        if os.getenv("WHEEL_POST_VERSION_STRATEGY_OVERRIDE", "None").upper() not in ("FALSE", "NONE"):
            post_version_strategy_override = os.getenv("WHEEL_POST_VERSION_STRATEGY_OVERRIDE")
        return post_version_strategy_override

    @property
    def dependencies_file(self) -> Path | None:
        return self.get("dependencies-file") and Path(self.get("dependencies-file")) or None

    @property
    def odoo_version_override(self) -> str | None:
        return self.get("odoo_version_override") and str(self.get("odoo_version_override")) or None

    @property
    def allow_no_odoo_version(self) -> bool:
        return self.get("allow_no_odoo_version", False)

    @property
    def addon_src(self) -> Path:
        return Path(self.get("src", "."))

    @property
    def metadata_src_manifest(self) -> bool:
        return self.get("metadata_src", "manifest") == "manifest"


class PyProjectConfig(PyProjectConfigBase):
    def __init__(self, config_dir: Path, project_config: dict[str, Any], tool_addon_odoo_wheel: dict[str, Any]) -> None:
        super().__init__(project_config)
        self.config_dir = config_dir
        self.addon_odoo_wheel = ToolConfig(tool_addon_odoo_wheel)

    @classmethod
    def from_config(cls, config_dir: Path, config: dict[str, Any]):
        project_config = config.get("project", {})
        return cls(config_dir, dict(project_config), dict(config.get("tool", {}).get("addon-odoo-wheel", {})))

    @property
    def name(self) -> str | None:
        return self.get("name")

    @property
    def version(self) -> str | None:
        return self.get("version")

    @property
    def dynamic(self) -> list[str]:
        return self.get("dynamic", [])

    @property
    def addon_src(self) -> Path:
        return self.config_dir / self.addon_odoo_wheel.addon_src


def _scm_ls_files(addon_dir: Path) -> list[str]:
    try:
        return subprocess.check_output(["git", "ls-files"], universal_newlines=True, cwd=addon_dir).strip().split("\n")
    except (subprocess.CalledProcessError, FileNotFoundError) as e:
        raise NoScmFound() from e


def copy_to(addon_dir: Path, dst: Path, pkg_exclude: Callable[[str, list[str]], list[str]]) -> None:
    if (Path(addon_dir) / "PKG-INFO").exists():
        # if PKG-INFO is present, assume we are in an sdist, copy everything
        shutil.copytree(addon_dir, dst, ignore=pkg_exclude)
        return
    # copy scm controlled files
    try:
        scm_files = _scm_ls_files(addon_dir)
    except NoScmFound:
        # NOTE This requires pip>=21.3 which builds in-tree. Previous pip versions
        # copied to a temporary directory with a different name than the addon, which
        # caused the resulting distribution name to be wrong.
        shutil.copytree(addon_dir, dst, ignore=pkg_exclude)
    else:
        dst.mkdir()
        for f in scm_files:
            d = Path(f).parent
            dstd = dst / d
            if not dstd.is_dir():
                dstd.mkdir(parents=True)
            shutil.copy(addon_dir / f, dstd)


def ensure_absent(paths: list[Path]) -> None:
    for path in paths:
        if path.exists():
            path.unlink()



import os
from pathlib import Path
from zipfile import ZipFile

import pytest
from manifestoo_core.git_postversion import POST_VERSION_STRATEGY_NONE
from whool.init import init_addon_dir

from addon_odoo_wheel.builder import build_wheel
from tests.utils import dir_changer


def test_build_wheel(module_data: Path, tmp_path: Path) -> None:
    with dir_changer(module_data / "module1"):
        wheel_name = build_wheel(os.fspath(tmp_path))
        assert wheel_name == "addon_odoo_module1-15.0.1.0.0-py3-none-any.whl"
        assert (tmp_path / wheel_name).exists()
        with ZipFile(tmp_path / wheel_name) as zf:
            names = zf.namelist()
            assert "odoo/addons/module1/__manifest__.py" in names
            assert "odoo/addons/module1/pyproject.toml" not in names


def test_build_wheel_without_scm(tmp_path: Path) -> None:
    addon_dir = tmp_path / "addon1"
    addon_dir.mkdir()
    addon_dir.joinpath("__manifest__.py").write_text("{'name': 'addon1', 'version': '16.0.1.0.0'}")
    addon_dir.joinpath("__init__.py").touch()
    init_addon_dir(addon_dir)
    wheel_path = tmp_path / "wheel"
    wheel_path.mkdir()
    with dir_changer(addon_dir):
        wheel_name = build_wheel(os.fspath(wheel_path))
        assert wheel_name == "addon_odoo_addon1-16.0.1.0.0-py3-none-any.whl"
        assert (wheel_path / wheel_name).exists()
        with ZipFile(wheel_path / wheel_name) as zf:
            names = zf.namelist()
            assert "odoo/addons/addon1/__manifest__.py" in names
            assert "odoo/addons/addon1/pyproject.toml" not in names


def test_build_wheel_git_missing(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    # Make sure we can't find git executable
    monkeypatch.setenv("PATH", "/foo")
    # Override strategy so Git is not needed
    monkeypatch.setenv("WHOOL_POST_VERSION_STRATEGY_OVERRIDE", POST_VERSION_STRATEGY_NONE)
    test_build_wheel_without_scm(tmp_path)



{
    "name": "Project1",
    "description": """Summary of Project1""",
    "author": "Author",
    "website": "https://website.fr",
    "category": "Technical",
    "version": "18.0.1.2.3",
    "license": "AGPL-3",
    "depends": ["base", "sale"],
    "installable": True,
    "application": False,
    "auto_install": False,
}



import os
from pathlib import Path

from addon_odoo_wheel.builder import (
    prepare_metadata_for_build_editable,
    prepare_metadata_for_build_wheel,
)
from tests.utils import dir_changer


def test_prepare_metadata_for_build_wheel(project2: Path, tmp_path: Path) -> None:
    with dir_changer(project2):
        dist_info_dir = prepare_metadata_for_build_wheel(os.fspath(tmp_path))
        assert dist_info_dir == "company_addon_project-18.0.1.2.3.dist-info"
        dist_info_path = tmp_path / dist_info_dir
        assert dist_info_path.is_dir()
        metadata_path = dist_info_path / "METADATA"
        assert metadata_path.is_file()


def test_prepare_metadata_for_build_editable(project2: Path, tmp_path: Path) -> None:
    with dir_changer(project2):
        dist_info_dir = prepare_metadata_for_build_editable(os.fspath(tmp_path))
        assert dist_info_dir == "company_addon_project-18.0.1.2.3.dist-info"
        dist_info_path = tmp_path / dist_info_dir
        assert dist_info_path.is_dir()
        metadata_path = dist_info_path / "METADATA"
        assert metadata_path.is_file()
